"""Utils to create the holes."""

from abc import ABC, abstractmethod
from typing import Tuple

import gmsh
import math


class Hole(ABC):
    """Abstract base class for holes.

    Attributes:
        center_x (float): x-coordinate of the center.
        center_y (float): y-coordinate of the center.
    """

    def __init__(self, center_x: float, center_y: float) -> None:
        """Initializes a Hole object."""
        self.center_x = center_x
        self.center_y = center_y

    @abstractmethod
    def perimeter(self):
        """Abstract method to calculate the perimeter of the hole."""
        pass

    @abstractmethod
    def to_dict(self):
        """Abstract method to convert the hole properties to a dictionary."""
        pass

    @abstractmethod
    def to_occ(self):
        """Abstract method to convert the hole to OpenCASCADE CAD."""
        pass

    @abstractmethod
    def get_hole_mesh_params(self):
        """Abstract method for hole mesh parameters."""
        pass


class CircularHole(Hole):
    """Circular hole.

    Attributes:
        radius (float): Hole radius.
    """

    def __init__(self, center_x: float, center_y: float, radius: float) -> None:
        """Initializes a CircularHole object."""
        super().__init__(center_x, center_y)
        self.radius = radius

    def perimeter(self) -> float:
        """Calculate the perimeter of the hole.

        Returns:
            float: The calculated perimeter.
        """
        return 2 * math.pi * self.radius

    def to_dict(self) -> dict:
        """Convert hole properties to a dictionary.

        Returns:
            dict: Hole properties.
        """
        return {
            "hole_type": "circular",
            "center_x": self.center_x,
            "center_y": self.center_y,
            "radius": self.radius
        }

    def to_occ(self) -> int:
        """Converts hole to OpenCASCADE CAD representation.

        Returns:
          hole_gmsh (int): The Gmsh entity ID representing the hole's
            OpenCASCADE CAD representation.
        """
        hole_gmsh = gmsh.model.occ.addDisk(xc=self.center_x,
                                           yc=self.center_y,
                                           zc=0,
                                           rx=self.radius,
                                           ry=self.radius)
        return hole_gmsh

    def get_hole_mesh_params(self) -> Tuple[float, float]:
        """Gets the mesh generation parameters for the hole.

        Metrics:
          - mesh_offset (float): Represents an offset for the boundaries of the
            holes, defining a region around the boundaries. Within this region,
            we have the ability to control the mesh elements size.
            The offset is equal to the radius.
          - predefined_element_size (float): Represents the predefined element
            size, defined as 1/4 of the perimeter.

        Returns:
            Tuple[float, float]: The mesh offset and the predefined element mesh
              size for the hole.
        """
        mesh_offset = self.radius
        predefined_element_size = self.perimeter() / 4

        return mesh_offset, predefined_element_size


class RectangularHole(Hole):
    """Rectangular hole.

    Attributes:
        half_size_x (float): Half size of the hole in the x-direction.
        half_size_y (float): Half size of the hole in the y-direction.
        angle (float): Positive angle of rotation in degrees around the hole
          center.
    """

    def __init__(self, center_x: float, center_y: float, half_size_x: float,
                 half_size_y: float, angle: float) -> None:
        """Initializes a RectangularHole object."""
        super().__init__(center_x, center_y)
        self.half_size_x = half_size_x
        self.half_size_y = half_size_y
        self.angle = angle

    def perimeter(self) -> float:
        """Calculate the perimeter of the hole.

        Returns:
            float: The calculated perimeter.
        """
        return self.half_size_x * 4 + self.half_size_y * 4

    def to_dict(self) -> dict:
        """Convert hole properties to a dictionary.

        Returns:
            dict: Hole properties.
        """
        return {
            "hole_type": "rectangular",
            "center_x": self.center_x,
            "center_y": self.center_y,
            "half_size_x": self.half_size_x,
            "half_size_y": self.half_size_y,
            "angle": self.angle
        }

    def to_occ(self) -> int:
        """Converts hole to OpenCASCADE CAD representation.

        Returns:
          hole_gmsh (int): The Gmsh entity ID representing the hole's
            OpenCASCADE CAD representation.
        """
        hole_gmsh = gmsh.model.occ.addRectangle(
            x=self.center_x - self.half_size_x,
            y=self.center_y - self.half_size_y,
            z=0,
            dx=self.half_size_x * 2,
            dy=self.half_size_y * 2)
        gmsh.model.occ.rotate(dimTags=[(2, hole_gmsh)],
                              x=self.center_x,
                              y=self.center_y,
                              z=0,
                              ax=0,
                              ay=0,
                              az=1,
                              angle=math.radians(self.angle))
        return hole_gmsh

    def get_hole_mesh_params(self) -> Tuple[float, float]:
        """Gets the mesh generation parameters for the hole.

        Metrics:
          - mesh_offset (float): Represents an offset for the boundaries of the
            holes, defining a region around the boundaries. Within this region,
            we have the ability to control the mesh elements size.
            The offset is equal to half of the minimum size of the hole.
          - predefined_element_size (float): Represents the predefined element
            size, defined as 1/4 of the perimeter.

        Returns:
            Tuple[float, float]: The mesh offset and the predefined element mesh
              size for the hole.
        """
        mesh_offset = min(self.half_size_x, self.half_size_y) / 2
        predefined_element_size = self.perimeter() / 4

        return mesh_offset, predefined_element_size


class EllipticalHole(Hole):
    """Elliptical hole.

    Attributes:
        semi_axis_x (float): The semi-axis along the x-direction.
        semi_axis_y (float): The semi-axis along the y-direction.
        angle (float): Positive angle of rotation in degrees around the hole
          center.
    """

    def __init__(self, center_x: float, center_y: float, semi_axis_x: float,
                 semi_axis_y: float, angle: float) -> None:
        """Initializes a EllipticalHole object."""
        super().__init__(center_x, center_y)
        self.semi_axis_x = semi_axis_x
        self.semi_axis_y = semi_axis_y
        self.angle = angle

    def perimeter(self) -> float:
        """Calculate the perimeter of the hole.

        Returns:
            float: The calculated perimeter.
        """
        return math.pi * (3 * (self.semi_axis_x + self.semi_axis_y) - math.sqrt(
            (3 * self.semi_axis_x + self.semi_axis_y) *
            (self.semi_axis_x + 3 * self.semi_axis_y)))

    def to_dict(self) -> dict:
        """Convert hole properties to a dictionary.

        Returns:
            dict: Hole properties.
        """

        return {
            "hole_type": "elliptical",
            "center_x": self.center_x,
            "center_y": self.center_y,
            "semi_axis_x": self.semi_axis_x,
            "semi_axis_y": self.semi_axis_y,
            "angle": self.angle
        }

    def to_occ(self) -> int:
        """Converts hole to OpenCASCADE CAD representation.

        Gmsh adheres to a standard where it expects the major axis
        (larger semi-axis) of an ellipse or disk to be oriented parallel to the
        X-axis, while the minor axis (smaller semi-axis) should align with the
        Y-axis.

        When the length of the semi-axis along the X-axis is less than the
        length of the semi-axis along the Y-axis, the code swaps the major and
        minor semi-axis values. This adjustment ensures that the shape aligns
        correctly with Gmsh's convention. Furthermore, a 90-degree rotation is
        applied to ensure the shape is properly oriented.

        Returns:
            hole_gmsh (int): The Gmsh entity ID representing the hole's
              OpenCASCADE CAD representation.
        """
        rx, ry = self.semi_axis_x, self.semi_axis_y
        angle = self.angle

        if self.semi_axis_x < self.semi_axis_y:
            rx, ry = self.semi_axis_y, self.semi_axis_x
            angle += 90

        hole_gmsh = gmsh.model.occ.addDisk(xc=self.center_x,
                                           yc=self.center_y,
                                           zc=0,
                                           rx=rx,
                                           ry=ry)
        gmsh.model.occ.rotate(dimTags=[(2, hole_gmsh)],
                              x=self.center_x,
                              y=self.center_y,
                              z=0,
                              ax=0,
                              ay=0,
                              az=1,
                              angle=math.radians(angle))

        return hole_gmsh

    def get_hole_mesh_params(self) -> Tuple[float, float]:
        """Gets the mesh generation parameters for the hole.

        Metrics:
          - mesh_offset (float): Represents an offset for the boundaries of the
            holes, defining a region around the boundaries. Within this region,
            we have the ability to control the mesh elements size.
            The offset is equal to the minimum value semi-axis.
          - predefined_element_size (float): Represents the predefined element
            size, defined as 1/4 of the perimeter.

        Returns:
            Tuple[float, float]: The mesh offset and the predefined element mesh
              size for the hole.
        """
        mesh_offset = min(self.semi_axis_x, self.semi_axis_y)
        predefined_element_size = self.perimeter() / 4

        return mesh_offset, predefined_element_size



"""Post-processing and visualization utilities of the heat sink scenario."""

import os
import pathlib
from typing import Literal, Sequence

import pyvista as pv

from inductiva.types import Path
from inductiva.utils import files


class HeatSinkOutput:
    """Heat sink simulation output."""

    def __init__(self, sim_output_path: Path):
        """Initializes a `HeatSinkOutput` object.

        Args:
            sim_output_path: Path to simulation output files.
        """

        self.sim_output_path = sim_output_path

    def render(
        self,
        movie_path: Path = "movie.mp4",
        fps: int = 10,
        cmap: str = "viridis",
        clim: Sequence[float] = (280, 300),
        view: Literal["isometric", "front", "rear", "top",
                      "side"] = "isometric",
    ):
        """Renders temporal evolution of the temperature.
        
        A movie is produced representing the temporal evolution of the
        temperature in the heat sink and in the surrounding air flow.
         
        The movie is saved in the `movie_path` location.

        Args:
            movie_path: Path to the movie file.
            fps: Number of frames per second to use in the movie.
            cmap: Colormap used to represent temperature.
            clim: Colorbar limits.
            view: View to use in the movie.
        """

        movie_path = files.resolve_path(movie_path)

        # The OpenFOAM data reader from PyVista requires that a file named
        # "foam.foam" exists in the simulation output directory.
        # Create this file if it does not exist.
        foam_file_path = os.path.join(self.sim_output_path, "foam.foam")
        pathlib.Path(foam_file_path).touch(exist_ok=True)

        reader = pv.OpenFOAMReader(foam_file_path)

        plotter = pv.Plotter(off_screen=True)

        # Set camera position for a nice view.
        if view == "isometric":
            plotter.view_vector([-0.67, 0.49, -0.58], viewup=[0.31, 0.90, 0.29])
            zoom_factor = 1.4
            slice_normal = (-1, 0, 0)
        elif view == "front":
            plotter.view_xy(negative=True)
            zoom_factor = 1.6
            slice_normal = (0, 0, -1)
        elif view == "rear":
            plotter.view_xy()
            zoom_factor = 1.6
            slice_normal = (0, 0, 1)
        elif view == "top":
            plotter.view_zx()
            zoom_factor = 1.2
            slice_normal = (0, -1, 0)
        elif view == "side":
            plotter.view_zy()
            zoom_factor = 1.2
            slice_normal = (-1, 0, 0)
        else:
            raise ValueError("Invalid view.")

        plotter.camera.zoom(zoom_factor)

        plotter.open_movie(movie_path, framerate=fps)

        for idx in range(1, reader.number_time_points):
            reader.set_active_time_point(idx)

            mesh = reader.read()

            fins_data = mesh["fins"]["boundary"]
            fluid_data = mesh["fluid"]["internalMesh"]

            fluid_data_slice = fluid_data.slice(normal=slice_normal,
                                                origin=(0, 0.02, 0))

            plotter.add_mesh(fins_data,
                             scalars="T",
                             cmap=cmap,
                             clim=clim,
                             show_scalar_bar=False)

            plotter.add_mesh(fluid_data_slice,
                             scalars="T",
                             cmap=cmap,
                             clim=clim,
                             show_scalar_bar=False)

            # Create custom color bar for temperature.
            plotter.add_scalar_bar(
                "Temperature [K]",
                position_x=0.45,
                width=0.5,
                title_font_size=18,
                label_font_size=18,
                fmt="%.0f",
                font_family="arial",
            )

            # Add time label.
            plotter.add_title(f"Time = {reader.active_time_value} [s]",
                              font_size=10)

            plotter.show_axes()

            plotter.write_frame()

        plotter.close()
